mat33 = tuple[float, float, float, float, float, float, float, float, float]
vec2 = tuple[float, float]
vec2i32 = tuple[int, int]
vec3 = tuple[float, float, float]


def mat33_new():
    return 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0


def vec2_new():
    return 0.0, 0.0


def vec3_new():
    return 0.0, 0.0, 0.0
